# -*- coding: utf-8 -*-
import wsgi


class HttpError(Exception): pass

class BasicAuth(Exception):
    def __init__(self):
        self.res = wsgi.Response('Authorization Error', 
            code=401, content_type='text/plain',
            headers={'WWW-Authenticate' : 'Basic realm="Enter password"'},
        )



# -*- coding: utf-8 -*-
import marshal
import base64

from pysi import cfg
from util import anticache_headers


cfg.set_defaults(
    BASIC_AUTH_PREFIXES = ('/', ),
    ANTICACHE_PREFIXES = ('/', ),
)

class BasicAuth(object):
    def process_request(self, rq):
        if rq.path.startswith(cfg.BASIC_AUTH_PREFIXES):
            rq.basic_user


class AntiCache(object):
    def process_response(self, rq, res):
        '''Browser anti-cache response middleware'''
        if rq.path.startswith(cfg.ANTICACHE_PREFIXES):
            res.headers.update(anticache_headers())


class Flash(object):
    def process_response(self, rq, res):
        '''Flash messages response middleware'''
        if rq._new_flashes:
            val = base64.urlsafe_b64encode(marshal.dumps(rq._new_flashes))
            res.set_cookie(cfg.FLASH_COOKIE_NAME, val)
        elif rq._clear_flashes:
            res.delete_cookie(cfg.FLASH_COOKIE_NAME)

    




# -*- coding: utf-8 -*-
import os

import wsgi
from config import cfg
from util import obj_from_str, cached_function

def render_to_string(rq, template, context=None):
    context = context or {}
    if not isinstance(context, dict):
        return unicode(context)
    for k, v in rq.context.iteritems():
        if k not in context:
            context[k] = v
    return _env().get_template(template).render(context or {})

def render(rq, to, context=None, **response_kwargs):
    if to == 'json':
        data = _json_dumps()(context)
        if 'callback' in rq.GET:
            data = '%s(%s)' % (rq.GET['callback'], data)  # jsonp
            content_type = 'application/x-javascript'
        else:
            content_type = 'application/json'
    elif to == 'text':
        data = unicode(context)
        content_type = 'text/plain'
    elif to == 'html':
        data = unicode(context)
        content_type = 'text/html'
    else:
        data = render_to_string(rq, to, context)
        content_type = 'text/html'
    if 'content_type' not in response_kwargs:
        response_kwargs['content_type'] = content_type
    return wsgi.Response(data, **response_kwargs)

@cached_function
def _json_dumps():
    try:
        import simplejson as json
    except ImportError:
        import json
    except ImportError:
        from django.utils import simplejson as json
    except ImportError:
        raise ImportError('Json module not found')
    return json.dumps

@cached_function
def _env():
    import jinja2
    return jinja2.Environment(
        loader = jinja2.FunctionLoader(obj_from_str(cfg.TEMPLATE_LOADER)),
        autoescape = True,
        cache_size = cfg.TEMPLATE_CACHE_SIZE,
        auto_reload = cfg.TEMPLATE_AUTO_RELOAD,
        extensions = cfg.JINJA2_EXTENSIONS,
    )

class FileLoader(object):
    mtimes = {}

    def load(self, path):
        try:
            f = open('templates/%s' % path)
        except IOError:
            try:
                app_name, path = path.split('/', 1)
                f = open('%s/templates/%s' % (app_name, path))
            except (IOError, ValueError):
                assert 0, 'template not found: %s' % path
        path = os.path.normpath(os.path.join(os.getcwd(), f.name))
        self.mtimes[path] = os.path.getmtime(path)
        return (f.read().decode('utf-8'), path, 
            lambda: self.mtimes[path] == os.path.getmtime(path))
file_loader = FileLoader().load



# -*- coding: utf-8 -*-
'''
Роутинг
'''
import urllib
import datetime

def main():
    # добавление
    m = Map()
    m.add_rule('/user/<int:id>/', 'user1', 'user')
    m.add_rule('/user/<int:id>/<str:name>/', 'user2', 'user')
    m.add_rule('/user/<*:slug>', 'user3', 'user')
    
    from om.pp import pp
    #~ pp(m.queue)
    #~ pp(m.rev)
    #~ return
    
    print m.find('/user/41/---/')
    print m.reverse('user', id=1)
    print m.reverse('user', id=1, name='pysi', x=4)
    print m.reverse('user', slug='ddd/bbb/sss/')


class Map(object):
    queue = []
    rev = {}
    charset = 'utf-8'
    url_prefix = None

    def add_rule(self, url, target, urlname=None):
        if self.url_prefix:
            url = '/%s/%s' % (self.url_prefix, url.lstrip('/'))
        match_funcs = []  # функции проверки папок пути
        path_size = None  # длина пути, -1 = бесконечный
        rev_chunks = ['']
        var_names = []
        for dir in url.lstrip('/ ').split('/'):
            func = get_match(dir)
            if func.__name__ == 'slug_dir':
                path_size = -1
            match_funcs.append(func)
            if dir.startswith('<') and dir.endswith('>'):
                var_type, var_name = dir[1:-1].split(':')
                rev_chunks.append(REVERSES[var_type](var_name))
                var_names.append(var_name)
            else:
                rev_chunks.append(dir)
        if path_size is None:
            path_size = len(match_funcs)
        self.queue.append((match_funcs, path_size, var_names, target))
        if urlname:
            self.rev.setdefault(urlname, []).append(
                ('/'.join(rev_chunks), var_names))
            if len(self.rev[urlname]) > 1:
                # сортируем, чтобы reverse использовал максимум переменных
                self.rev[urlname].sort(key=lambda x: -len(x[1]))

    def find(self, url):
        dirs = url.lstrip('/ ').split('/')
        for match_funcs, path_size, var_names, target in self.queue:
            if path_size == -1 or path_size == len(dirs):
                vals = []
                for i, func in enumerate(match_funcs):
                    match, val = func(dirs[i])
                    if match:
                        if val is not None:
                            if val == '<//__slug__//>':
                                vals.append('/'.join(dirs[i:]))
                            else:
                                vals.append(val)
                    else:
                        break
                else:
                    return target, dict(zip(var_names, vals)) if vals else {}
        return None, None

    def reverse(self, urlname, **kwargs):
        try:
            items = self.rev[urlname]
        except KeyError:
            assert 0, 'Could not find url for "%s"' % urlname
        # выбераем с максимальным количеством используемых переменных
        for tpl, vars in items:
            try:
                url = urllib.quote(tpl % kwargs)
            except (TypeError, KeyError):
                pass
            else:
                # GET параметры
                if len(kwargs) > len(vars):
                    ext_vars = {}
                    for k, v in kwargs.iteritems():
                        if k not in vars:
                            if isinstance(v, unicode):
                                v = v.encode(self.charset)
                            ext_vars[k] = v
                    if ext_vars:
                        url += '?' + urllib.urlencode(ext_vars)
                return url
        assert 0, 'Could not find url for "%s" witch kwargs: %s' % (urlname,
            ', '.join('%s %s' % (k, type(v)) for k, v in kwargs.iteritems()))
            
def match_const(var):
    def const_dir(dir):
        return dir == var, None
    return const_dir

def match_str(var):
    def str_dir(dir):
        return True, urllib.unquote_plus(dir)
    return str_dir
    
def match_int(var):
    def int_dir(dir):
        if dir.isdigit():
            return True, int(dir)
        else:
            return False, None
    return int_dir

def match_date(var):
    def data_dir(dir):
        try:
            dt = datetime.datetime.strptime(dir, '%Y-%m-%d')
        except:
            return False, None
        return True, datetime.date(dt.year, dt.month, dt.day)
    return data_dir

def match_slug(var):
    def slug_dir(dir):
        return (True, '<//__slug__//>')
    return slug_dir


MATCHS = {
    'int': match_int,
    'str': match_str,
    'date': match_date,
    '*': match_slug,
}

def get_match(dir):
    if dir.startswith('<') and dir.endswith('>'):
        name = dir[1:-1]
        name, var = name.split(':')
        func = MATCHS.get(name)
        if func:
            return func(dir)
    return match_const(dir)

REVERSES = {
    'int': lambda var: '%%(%s)i' % var,
    'str': lambda var: '%%(%s)s' % var,
    'date': lambda var: '%%(%s)s' % var,
    '*': lambda var: '%%(%s)s' % var,
}


if __name__ == '__main__':
    main()
else:
    #######################
    ### Привязка к pysi ###
    #######################
    from config import cfg
    import wsgi


    urls = Map()
    urls.charset = cfg.CHARSET

    def auto_routing(rq):
        '''
        Автоматический роутинг
        '''
        func, kwargs = urls.find(rq.path)
        if func is None:
            if not rq.path.endswith('/'):
                # забыли закрывающий слеш?
                func, kwargs = urls.find(rq.path + '/')
                if func:
                    return wsgi.redirect(rq.path + '/')        
            wsgi.abort(404)
        return func(rq, **kwargs)
        
    def add_urls(*rules):
        ''' Добавление правил списком'''
        for rule in rules:
            urls.add_rule(*rule)
            
    def url4(*args, **kwargs):
        return urls.reverse(*args, **kwargs)
        
    def add_apps(apps):
        '''
        Добавление приложений
            apps = ['mod1', ('mod2', url_prefix), 'mod3']
        '''
        for mod in apps:
            if isinstance(mod, tuple):
                mod, prefix = mod
                urls.url_prefix = prefix.strip().strip('/') or None
            app_name = '%s.%s' % (mod, 'views')
            __import__(app_name)
            urls.url_prefix = None
    


# -*- coding: utf-8 -*-
import wsgi


class Cfg(object):
    DEBUG = False
    BASIC_AUTH_SSL_ONLY = False  # basic-авторизация только по https
    BASIC_AUTH_DB = {}
    INSTALLED_APPS = []
    CONTEXT_PROCESSORS = []
    REQUEST_MIDDLEWARES = []
    RESPONSE_MIDDLEWARES = []
    MIDDLEWARES = []
    CHARSET = 'utf-8'
    POST_MAX_SIZE = 1024 * 1024 * 1  # максимальный размер пост-данных, None - не ограничивать
    POST_MAX_MEMFILE = 1024 * 100  # максимальный размер POST-данных, загружаемый в память. Если больше, создаётся временный файл.
    ERROR_PAGES = {}
    TEMPLATE_CACHE_SIZE = 250
    TEMPLATE_AUTO_RELOAD = False
    TEMPLATE_LOADER = 'pysi.template.file_loader'
    FLASH_COOKIE_NAME = u'pysi.flash'
    ROUTING_QUEUE = ['pysi.auto_routing']  # очередь функций роутинга
    JINJA2_EXTENSIONS = ['jinja2.ext.with_']

    def set_defaults(self, **kwargs):
        '''
        Добавляем дефолтные атрибуты
        '''
        for k, v in kwargs.iteritems():
            if not hasattr(self, k) and k.isupper():
                setattr(self, k, v)

cfg = Cfg()



# -*- coding: utf-8 -*-
'''
Сигналы

    @pysi.signal.connect('sig')
    def my_sig(rq, var):
        rq.context['test'] = var

    @pysi.view('/')
    def index(rq):
        pysi.signal.send('sig', rq, u'сигналит')
        return {'a': 1}
'''

_signals = {}

def connect(signal_name):
    def decorator(func):
        _signals.setdefault(signal_name, []).append(func)
        return func
    return decorator
    
def send(signal_name, *args, **kwargs):
    if signal_name in _signals:
        for func in _signals[signal_name]:
            func(*args, **kwargs)